#include <stdlib.h>
#include <stdio.h>


void ld_mat_ijv(char *file_name, double *A, int m, int n, int load_only_upper){

	int t,i,k;
    FILE *f;
    int mtx_size,band,nnz;
    int ii,jj;
	double vv;

    f=fopen(file_name,"r");
    if(!f){
       printf("Couldn't open file %d:%s\n",__LINE__,__FILE__);
       exit(1);
    }

	fscanf(f,"%d%d%d",&mtx_size,&band,&nnz);
	if(mtx_size!=n || band*6+1!=m){
	   printf("Error: Expected to %dx%d matrix, but %s is a %dx%d matrix.\n",m,n,file_name,band*6+1,mtx_size);
       fclose(f);
	   exit(1);
	}

	t=m*n;
	for(i=0; i<t; i++){
		A[i]=0.0;
	}

	for(k=0; k<nnz; k++){
       fscanf(f,"%d%d%lf",&ii,&jj,&vv);
	   if(jj<ii){
		   if(!load_only_upper){
  		      *(A+band*3+m*(jj-1)+(ii-jj))=vv;
		   }
	   }
	   else{
		   *(A+band*3+m*(jj-1)-(jj-ii))=vv;
	   }
	}
   
	fclose(f);
}


void ld_mat(char *file_name, double *A, int m, int n){
   
   int i,j;
   FILE *f; 
   long c=0,k=0;
   char str[100];
   double *B;
   int num_of_files,num_raws_per_file,num_raws_last_file;

   f=fopen(file_name,"r");
   if(!f){
      printf("Couldn't open file %d:%s\n",__LINE__,__FILE__);
      exit(1);
   }
   fscanf(f,"%d%d%d",&num_of_files,&num_raws_per_file,&num_raws_last_file);
   fclose(f);

   if(m!=(num_of_files-1)*num_raws_per_file+num_raws_last_file){
      printf("Matrix size in %s (%d) is different than what expected (%d).\n",
	  file_name,(num_of_files-1)*num_raws_per_file+num_raws_last_file,m);
      exit(1);
   }

   B=A;
   for(k=0; k<num_of_files-1; k++){
	  sprintf(str,"%s.%d",file_name,k);
      f=fopen(str,"r");
      if(!f){
         printf("Couldn't open file.\n");
         exit(1);
      }
      for(i=0; i<num_raws_per_file; i++){
         for(j=0; j<n; j++){
  		    fscanf(f,"%lf", &(B[i+m*j]));
		 }
	  }
      fclose(f);
	  B+=num_raws_per_file/**n*/;
   }
   sprintf(str,"%s.%d",file_name,k);
   f=fopen(str,"r");
   for(i=0; i<num_raws_last_file; i++){
      for(j=0; j<n; j++){
	    fscanf(f,"%lf", &(B[i+m*j]));
	  }
   }


   
   fclose(f);
}


void st_mat(char *file_name, double *A, int m, int n){

   int i,j;
   FILE *f; 
   long c=0,k=0;
   char str[100];

   sprintf(str,"%s",file_name);
   f=fopen(str,"w");

   if(!f){
      printf("Couldn't open file\n");
      exit(1);
   }

   for(i=0; i<m; i++){
	   for(j=0; j<n; j++){
		  fprintf(f,"%.16le ", A[i+m*j]);
	  }
      fprintf(f,"\n");
   }

   fclose(f);
   
}

