
//#define MY_DEBUG
#ifndef _GROWTH_
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#endif

#define BAND_FACTOR 3
#define EPSILON 1e-250

#define ab_ref(a_1,a_2) ab[(a_2)*ab_dim1 + a_1]
#define max(a,b) ((a) >= (b) ? (a) : (b))
#define abs(x) ((x) >= 0 ? (x) : -(x))


int xerbla_(char *srname, int *info);
int idamax_(int *n, double *dx, int *incx);
int dsyr_(char *uplo, int *n, double *alpha, double *x, int *incx, double *a, int *lda);
int dscal_(int *n, double *da, double *dx, int *incx);
int dswap_(int *n, double *dx, int *incx, double *dy, int *incy);
int daxpy_(int *n, double *da, double *dx, int *incx, double *dy, int *incy);
int dcopy_(int *n, double *dx, int *incx, double *dy, int *incy);
void prn_bnd_mtrx(int n, double *A, int lda);
void st_mat(char *file_name, double *A, int m, int n);


#ifdef _GROWTH_
int growth_snapback (char *uplo, /*int*/long *n, /*int*/ long *kd, double *ab, /*int*/long *ldab, int *bandwidth_u, int *bandwidth_l, int *step_type, double *workspace, /*int*/long *info){
#else
int snapback (char *uplo, /*int*/long *n, /*int*/ long *kd, double *ab, /*int*/long *ldab, int *bandwidth_u, int *bandwidth_l, int *step_type, double *workspace, /*int*/long *info){
#endif
   int i,j;
   double aii,amm,ami;
   double d,d1;
   int ab_dim1,ab_offset;
   int n1,kd1;
   int ldv1,ldv2;
   int ip1,ip2,ip3;
   int bw, new_bw;
   int tmp;
   double th1=1.0/3, th2=1.0; 
   double norm, sin_alpha, cos_alpha, cos_alpha1, givens1, givens2;
   int subdiag_beg;

//   double workspace[2000];
 

   ab_dim1 = *ldab;
   ab_offset = 1 + ab_dim1 * 1;
   ab -= ab_offset;
   bandwidth_u--;
   bandwidth_l--;
   step_type--;

   n1=*n;
   kd1=*kd;

   if(*ldab<(BAND_FACTOR*2)*kd1+1){
      *info=2;
      return 2;
   }

   subdiag_beg=kd1*BAND_FACTOR;

   for(i=1; i<=n1; i++)
      bandwidth_u[i]=kd1;

   for(i=n1; i>n1-kd1; i--)
      bandwidth_u[i]=n1-i;


   //meanwhile supports only uplo='u':
   if(uplo[0]=='u') {

     ldv1=*ldab-1;
     ldv2=1;

     i=1;
     while(i<=n1){
#ifdef _GROWTH_

        calc_reduced_max(&ab_ref(subdiag_beg+1,i),&(bandwidth_u[i]),n1-i+1,*ldab,0);
        calc_max_band(&(bandwidth_u[i]), &(bandwidth_l[i]), n1-i+1);
#endif

        bw=bandwidth_u[i];

        aii=ab_ref(subdiag_beg+1,i);
        //find max in the i'th row (let's call its index k):
        tmp=bw+1;
        ip1=i-1+idamax_(&tmp,&(ab_ref(subdiag_beg+1,i)),&ldv1);
        //find max in k'th column:
        tmp=ip1-i+1;
        ip2=subdiag_beg-(ip1-i)+idamax_(&tmp,&(ab_ref(subdiag_beg+1-(ip1-i),ip1)),&ldv2);
        //find max in k'th row:
        tmp=bandwidth_u[ip1]+1;
        ip3=ip1-1+idamax_(&tmp,&(ab_ref(subdiag_beg+1,ip1)),&ldv1);
        amm=max(abs(ab_ref(ip2,ip1)),abs(ab_ref(subdiag_beg+1-(ip3-ip1),ip3)));
        ami=abs(ab_ref(subdiag_beg+1-(ip1-i),ip1));

        for(j=i+1; j<=i+bw; j++)
           if(bandwidth_u[j]<bw-(j-i))
              bandwidth_u[j]=bw-(j-i);

        if(ami<EPSILON){
           step_type[i]=4;
           i++;
           continue;
        }


        //Condition for an LDL^T step - This is BK condition. the second condition
        //should be eliminated in order to have bounded factors.
        if( abs(aii) >= th1*ami || abs(aii)*amm >= th1*ami*ami || ip1==i){
            d=-1/aii;
            dsyr_("Upper", &bw, &d, &ab_ref(subdiag_beg,i+1),&ldv1,
               &ab_ref(subdiag_beg+1,i+1), &ldv1);
            d=1/aii;
            dscal_(&bw, &d, &ab_ref(subdiag_beg,i+1), &ldv1);
            step_type[i]=0;
            i++;

        // in case LDL^T step is not stable:
        }else{

           bandwidth_l[i]=bw;

           for(j=1; j<bw; j++){

              if(abs(ab_ref(subdiag_beg+1-j,i+j))<EPSILON)
                 ab_ref(subdiag_beg+1-j,i+j)=0;
              if(abs(ab_ref(subdiag_beg-j,i+j+1))<EPSILON)
                 ab_ref(subdiag_beg-j,i+j+1)=0;

              ab_ref(subdiag_beg+2,i+j)=ab_ref(subdiag_beg,i+j+1);

              {
                 double la=ab_ref(subdiag_beg-j,i+j+1);
                 double lb=ab_ref(subdiag_beg+1-j,i+j);
                 drotg_(&la,&lb,&cos_alpha,&sin_alpha);
              }
              cos_alpha=(sin_alpha<0?-1:1)*cos_alpha;
              sin_alpha=abs(sin_alpha);
              tmp=max(bandwidth_u[i+j+1]+2,bandwidth_u[i+j]+1);
              drot_(&tmp,&ab_ref(subdiag_beg+2,i+j),&ldv1,
                         &ab_ref(subdiag_beg+1,i+j),&ldv1,&cos_alpha,&sin_alpha);
              tmp=j+2;

              drot_(&tmp,&ab_ref(subdiag_beg-j,i+j+1),&ldv2,
                         &ab_ref(subdiag_beg-j+1,i+j),&ldv2,&cos_alpha,&sin_alpha);

              //tmp=bandwidth_u[i+j+1];
              bandwidth_u[i+j+1]=max(bandwidth_u[i+j+1],bandwidth_u[i+j]-1);
              bandwidth_u[i+j]=bandwidth_u[i+j+1]+1;
              //tmp=bandwidth_u[i+j+1];
              //bandwidth_u[i+j+1]=max(bandwidth_u[i+j]-1,0);
              //bandwidth_u[i+j]=max(tmp+1,bandwidth_u[i+j+1]+1);
              if(cos_alpha==0)
                 ab_ref(subdiag_beg+1+j,i)=1;
              else if(sin_alpha<abs(cos_alpha))
                 ab_ref(subdiag_beg+1+j,i)=(cos_alpha>0?1:-1)*0.5*sin_alpha;
              else              
                 ab_ref(subdiag_beg+1+j,i)=2/cos_alpha;


           }

           givens1=ab_ref(subdiag_beg+1-bw,i+bw);
           givens2=ab_ref(subdiag_beg+1,i);
           norm=sqrt(givens1*givens1+givens2*givens2);
           cos_alpha1=(givens1>0?1:-1)*givens2/norm;
           sin_alpha=abs(givens1/norm);

           if(abs(cos_alpha1)<EPSILON)
              ab_ref(subdiag_beg+1+bw,i)=1;
           else if(abs(sin_alpha)<abs(cos_alpha1))
              ab_ref(subdiag_beg+1+bw,i)=0.5*(abs(cos_alpha1)>0?1:-1)*sin_alpha;
           else              
              ab_ref(subdiag_beg+1+bw,i)=2*(abs(sin_alpha)>0?1:-1)/cos_alpha1;
           
           //find max in (i+bw)'th column:
           tmp=bw+1;
           ip2=subdiag_beg-bw+idamax_(&tmp,&(ab_ref(subdiag_beg+1-bw,i+bw)),&ldv2);
           //find max in k'th row:
           tmp=bandwidth_u[i+bw]+1;
           ip3=i+bw-1+idamax_(&tmp,&(ab_ref(subdiag_beg+1,i+bw)),&ldv1);
           amm=max(abs(ab_ref(ip2,i+bw)),abs(ab_ref(subdiag_beg+1-(ip3-i-bw),ip3)));
           aii=-ab_ref(subdiag_beg+1-bw,i+bw)*sin_alpha+ab_ref(subdiag_beg+1,i+bw)*cos_alpha1;

           //Update i'th row:
           ab_ref(subdiag_beg+1,i)=ab_ref(subdiag_beg+1,i)*cos_alpha1
                                      +ab_ref(subdiag_beg+1-bw,i+bw)*sin_alpha;
           ab_ref(subdiag_beg+1-bw,i+bw)=ab_ref(subdiag_beg+1-bw,i+bw)*cos_alpha1;

           d=sin_alpha;
		   daxpy_(&bw,&d,&ab_ref(subdiag_beg-bw+2,i+bw),&ldv2,
                        &ab_ref(subdiag_beg,i+1),&ldv1);
           //(assumes ab has zeros outside its band:)
           daxpy_(&bandwidth_u[i+bw],&d,&ab_ref(subdiag_beg,i+bw+1),&ldv1,
                          &ab_ref(subdiag_beg-bw,i+bw+1),&ldv1);
           bandwidth_u[i]=bw+bandwidth_u[i+bw];


           //Decide whether or not to use our new transformation:
           if(abs(amm*cos_alpha1) >= th2*abs(aii)){
              ab_ref(subdiag_beg+1,i+bw)=aii/cos_alpha1;
              step_type[i]=1;
              i++;
           }else{
              ab_ref(subdiag_beg+1,i+bw)=aii;

#ifndef MY_DEBUG
              workspace[0]=ab_ref(subdiag_beg+1,i+bw);
              tmp=bw-1;
              dcopy_(&tmp,&ab_ref(subdiag_beg+2-bw,i+bw),&ldv2,
                          workspace+1,&ldv2);
              tmp=bandwidth_u[i+bw];
              dcopy_(&tmp,&ab_ref(subdiag_beg,i+bw+1),&ldv1,
                          workspace+bw,&ldv2);

              for(j=bw-1;j>0;j--){

//                 tmp=bandwidth_u[i+j]-1;
                 tmp=max(bandwidth_u[i+j]-(bw-j),bandwidth_u[i+j+1]+1-(bw-j));
                 dcopy_(&tmp,&ab_ref(subdiag_beg-(bw-j),i+bw+1),&ldv1,
                             &ab_ref(subdiag_beg-(bw-1-j),i+bw+1),&ldv1);
                 //dcopy_(&tmp,&ab_ref(subdiag_beg-1,i+j+2),&ldv1,
                 //            &ab_ref(subdiag_beg,i+j+2),&ldv1);
                 tmp=j;
                 dcopy_(&tmp,&ab_ref(subdiag_beg+2-j,i+j),&ldv2,
                           &ab_ref(subdiag_beg+2-j,i+j+1),&ldv2);

                 //ab_ref(subdiag_beg,i+j+1)=ab_ref(subdiag_beg,i+j+1);

              }

              for(j=bw-1;j>0;j--){
                 tmp=bandwidth_u[i+j+1];
                 bandwidth_u[i+j+1]=max(bandwidth_u[i+j]-1,0);
                 bandwidth_u[i+j]=tmp+1;
              }

              tmp=bandwidth_u[i+1]+1;
              dcopy_(&tmp,workspace,&ldv2,
                          &ab_ref(subdiag_beg+1,i+1),&ldv1);

#else
// Old version:
              for(j=bw-1;j>0;j--){
                 ab_ref(subdiag_beg+2,i+j)=ab_ref(subdiag_beg,i+j+1);
                 tmp=max(bandwidth_u[i+j+1]+2,bandwidth_u[i+j]+1);
                 dswap_(&tmp,&ab_ref(subdiag_beg+1,i+j),&ldv1,
                             &ab_ref(subdiag_beg+2,i+j),&ldv1);
                 tmp=j+1;
                 dswap_(&tmp,&ab_ref(subdiag_beg+2-j,i+j),&ldv2,
                           &ab_ref(subdiag_beg+1-j,i+j+1),&ldv2);
                 tmp=bandwidth_u[i+j+1];
                 bandwidth_u[i+j+1]=max(bandwidth_u[i+j]-1,0);
                 bandwidth_u[i+j]=tmp+1;
              }

#endif

              step_type[i]=2;
              i++;


              for(j=1; j<bw-1; j++){

                 if(abs(ab_ref(subdiag_beg+1-j,i+j))<EPSILON)
                    ab_ref(subdiag_beg+1-j,i+j)=0;
                 if(abs(ab_ref(subdiag_beg-j,i+j+1))<EPSILON)
                    ab_ref(subdiag_beg-j,i+j+1)=0;

                 ab_ref(subdiag_beg+2,i+j)=ab_ref(subdiag_beg,i+j+1);





                 {
                    double la=ab_ref(subdiag_beg-j,i+j+1);
                    double lb=ab_ref(subdiag_beg+1-j,i+j);
                    drotg_(&la,&lb,&cos_alpha,&sin_alpha);
                 }
                 cos_alpha=(sin_alpha<0?-1:1)*cos_alpha;
                 sin_alpha=abs(sin_alpha);
                 tmp=max(bandwidth_u[i+j+1]+2,bandwidth_u[i+j]+1);
                 drot_(&tmp,&ab_ref(subdiag_beg+2,i+j),&ldv1,
                            &ab_ref(subdiag_beg+1,i+j),&ldv1,&cos_alpha,&sin_alpha);
                 tmp=j+2;

                 drot_(&tmp,&ab_ref(subdiag_beg-j,i+j+1),&ldv2,
                            &ab_ref(subdiag_beg-j+1,i+j),&ldv2,&cos_alpha,&sin_alpha);


                 bandwidth_u[i+j+1]=max(bandwidth_u[i+j+1],bandwidth_u[i+j]-1);
                 bandwidth_u[i+j]=bandwidth_u[i+j+1]+1;
                 if(cos_alpha==0)
                    ab_ref(subdiag_beg+1-j,i+j)=1;
                 else if(sin_alpha<abs(cos_alpha))
                    ab_ref(subdiag_beg+1-j,i+j)=(cos_alpha>0?1:-1)*0.5*sin_alpha;
                 else              
                    ab_ref(subdiag_beg+1-j,i+j)=2/cos_alpha;

              }


              d=-cos_alpha1/ab_ref(subdiag_beg+1,i);

              if(bw>1)
                  bw-=1;


              new_bw=1+bandwidth_u[i]-bw;
  
              for(j=i+bw; j<=i+bandwidth_u[i]; j++)
                 if(bandwidth_u[j]<bandwidth_u[i]-(j-i))
                    bandwidth_u[j]=bandwidth_u[i]-(j-i);

              dsyr_("Upper", &new_bw, &d, &ab_ref(subdiag_beg-bw+1,i+bw),&ldv1,
                  &ab_ref(subdiag_beg+1,i+bw), &ldv1);
//              d=(-1)*d/cos_alpha1;
              d=1/ab_ref(subdiag_beg+1,i);
              dscal_(&new_bw, &d, &ab_ref(subdiag_beg-bw+1,i+bw), &ldv1);

              step_type[i]=3;
              i++;

           }
         }
         
      }
   }else{
      *info=1;
      return 1;
   }


#ifdef _GROWTH_
        growth_factor=mx/first_mx;
        mx=0;
        calc_reduced_max(&ab_ref(subdiag_beg+1,1),bandwidth_u+1,n1,*ldab,1);
#endif

   *info=0;
   return 0;
}

